from . import api, client, enums, exceptions, models, util
from .client import Client
from ._metadata import __version__, __title__, __author__, __license__, __description__

__all__ = "api", "client", "enums", "exceptions", "models", "util", "Client"


from datetime import datetime
from enum import Enum
from typing import Type

from backoff import on_exception, expo
from httpx import HTTPStatusError

from pymondis.exceptions import NoEnumMatchError

default_backoff = on_exception(
    expo,
    HTTPStatusError,
    max_tries=3,
    giveup=lambda status: 400 <= status.response.status_code < 500
)


def get_enum_element(enum: Type[Enum], value: str) -> Enum:
    """Zamienia string-a na element enum-a"""
    for element in enum:
        if element.value == value:
            return element
    else:
        raise NoEnumMatchError(enum, value)


def date_converter(value: str | datetime) -> datetime:
    """Zamienia string-a na datetime"""
    return value if isinstance(value, datetime) else datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")


def character_converter(string: str) -> str | None:
    """
    Zamienia 'Nazwa postaci Quatromondis' na None,
    bo ktoś stwierdził że taka będzie wartość jak ktoś nie ma nazwy...
    """
    return None if string == "Nazwa postaci Quatromondis" else string


def empty_string_converter(string: str) -> str | None:
    """Zamienia pustego string-a na None"""
    return string if string else None


def enum_converter(enum: Type[Enum]):
    """Wrapper get_enum_element"""

    def inner_enum_converter(value: str | Enum) -> Enum:
        if isinstance(value, Enum):
            return value
        return get_enum_element(enum, value)

    return inner_enum_converter



from datetime import datetime
from typing import List, Dict

from httpx import AsyncClient

from .util import default_backoff
from ._metadata import __version__, __title__

class HTTPClient(AsyncClient):
    def __init__(
            self,
            timeout: float | None = None,
            *,
            base_url: str = "https://quatromondisapi.azurewebsites.net/api"
    ):
        super().__init__(timeout=timeout)
        self.base: str = base_url
        self.headers = {"User-Agent": "{}/{}".format(__title__, __version__)}

    @default_backoff
    async def get_resource(
            self,
            url: str,
            cache_time: datetime | None = None,
            cache_content: bytes | None = None
    ) -> bytes:
        headers = {
            "If-Modified-Since": cache_time.strftime("%a, %d %b %Y %H:%M:%S GMT")} if cache_time is not None else {}
        response = await self.get(
            url,
            headers=headers
        )
        if response.status_code == 304:
            return cache_content
        response.raise_for_status()
        return response.content

    @default_backoff
    async def get_camps(self) -> List[dict]:
        response = await self.get(
            self.base + "/Camps",
            headers={"Accept": "application/json"}
        )
        response.raise_for_status()
        return response.json()

    @default_backoff
    async def post_events_inauguration(self, reservation_model: dict):
        response = await self.post(
            self.base + "/Events/Inauguration",
            json=reservation_model
        )
        response.raise_for_status()

    @default_backoff
    async def get_images_galleries_castle(self, castle: str) -> List[Dict[str, str | int | bool]]:
        response = await self.get(
            self.base + "/Images/Galeries/Castle/{}".format(castle),  # 'Galeries' XD
            headers={"Accept": "application/json"})
        response.raise_for_status()
        return response.json()

    @default_backoff
    async def get_images_galleries(self, gallery_id: int) -> List[Dict[str, str]]:
        response = await self.get(
            self.base + "/Images/Galeries/{}".format(gallery_id),  # Znowu 'Galeries'
            headers={"Accept": "application/json"})
        response.raise_for_status()
        return response.json()

    @default_backoff
    async def post_orders_four_worlds_beginning(self, purchaser: dict):
        response = await self.post(
            self.base + "/Orders/FourWorldsBeginning",
            json=purchaser
        )
        response.raise_for_status()

    @default_backoff
    async def post_parents_zone_survey(self, survey_hash: str, result: dict):
        response = await self.post(
            self.base + "/ParentsZone/Survey/{}".format(survey_hash),
            json=result
        )
        response.raise_for_status()

    @default_backoff
    async def get_parents_zone_crew(self) -> List[dict]:
        response = await self.get(
            self.base + "/ParentsZone/Crew",
            headers={"Accept": "application/json"}
        )
        response.raise_for_status()
        return response.json()

    @default_backoff
    async def post_parents_zone_apply(self):
        raise NotImplementedError(
            "Nie mogę aktualnie zaimplementować tej metody, bo nie wiem jak są wysyłane dane"
            "Jeśli jest ci potrzebna możesz otworzyć nowy issue: https://github.com/Asapros/pymondis/issues"
        )
        # Dane najprawdopodobniej są wysyłane jako form, ale nie ma tego w swagger-ze, a ja jestem borowikiem w
        # javascript-a i nie czaje o co chodzi, dodajcie do dokumentacji pls

    @default_backoff
    async def post_reservations_subscribe(self, reservation_model: dict) -> List[str]:
        response = await self.post(
            self.base + "/Reservations/Subscribe",
            json=reservation_model,
            headers={"Accept": "application/json"}
        )
        response.raise_for_status()
        return response.json()

    @default_backoff
    async def post_reservations_manage(self, pri: dict) -> Dict[str, str | bool]:
        response = await self.post(
            self.base + "/Reservations/Manage",
            json=pri,
            headers={"Accept": "application/json"}
        )
        response.raise_for_status()
        return response.json()

    @default_backoff
    async def patch_vote(self, category: str, name: str):
        response = await self.patch(  # A mnie dalej zastanawia czemu tu patch jest, a nie post...
            self.base + "/Vote/{}/{}".format(category, name)
        )
        response.raise_for_status()

    @default_backoff
    async def get_vote_plebiscite(self, year: int) -> List[Dict[str, str | int | bool]]:
        response = await self.get(
            self.base + "/Vote/plebiscite/{}".format(year),
            # Jedyny endpoint gdzie słowo w ścieżce nie się zaczyna dużą literą...
            headers={"Accept": "application/json"}
        )
        response.raise_for_status()
        return response.json()

    async def __aenter__(self) -> "HTTPClient":  # Type-hinting
        await super().__aenter__()
        return self



from enum import Enum
from typing import Type


class NotFullyImplementedWarning(UserWarning):
    pass


class RevoteError(Exception):
    def __init__(self, category: str):
        super().__init__("Próbowałeś zagłosować na kategorię '{}' drugi raz".format(category))


class NoEnumMatchError(Exception):
    def __init__(self, enum: Type[Enum], value):
        super().__init__("Nie znaleziono żadnych pasujących elementów w {} dla wartości: {}".format(enum, value))



from typing import List


from .api import HTTPClient
from .enums import Castle
from .models import (
    Camp,
    CrewMember,
    Purchaser,
    Gallery,
    EventReservationSummary,
    PlebisciteCandidate,
    WebReservationModel,
    ParentSurveyResult
)


class Client:
    def __init__(self, http: HTTPClient | None = None):
        self.http: HTTPClient = HTTPClient() if http is None else http

    async def get_camps(self) -> List[Camp]:
        camps = await self.http.get_camps()
        return [Camp.init_from_dict(camp) for camp in camps]

    async def reserve_inauguration(self, reservation: EventReservationSummary):
        await self.http.post_events_inauguration(reservation.to_dict())

    async def get_galleries(self, castle: Castle) -> List[Gallery]:
        galleries = await self.http.get_images_galleries_castle(castle.value)
        return [Gallery.init_from_dict(gallery, http=self.http) for gallery in galleries]

    async def order_fwb(self, purchaser: Purchaser):
        await self.http.post_orders_four_worlds_beginning(purchaser.to_dict())

    async def submit_survey(self, survey_hash: str, result: ParentSurveyResult):
        raise NotImplementedError(
            "Ta metoda będzie brała ParentSurveyResult jako drugi argument"
            "Jeśli chcesz pomóc w jej implementacji otwórz nowy issue: https://github.com/Asapros/pymondis/issues"
        )

    async def get_crew(self) -> List[CrewMember]:
        crew = await self.http.get_parents_zone_crew()
        return [CrewMember.init_from_dict(crew_member, http=self.http) for crew_member in crew]

    async def apply_for_job(self):
        await self.http.post_parents_zone_apply()

    async def reserve_camp(self, reservation: WebReservationModel) -> List[str]:
        codes = await self.http.post_reservations_subscribe(reservation.to_dict())
        return codes

    async def get_plebiscite(self, year: int) -> List[PlebisciteCandidate]:
        candidates = await self.http.get_vote_plebiscite(year)
        return [PlebisciteCandidate.init_from_dict(candidate, http=self.http) for candidate in candidates]

    async def __aenter__(self) -> "Client":
        await self.http.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.http.__aexit__(exc_type, exc_val, exc_tb)


from datetime import datetime
from typing import List
from warnings import warn

from attr import attrib, attrs
from attr.validators import instance_of, optional as v_optional, deep_iterable
from attr.converters import optional as c_optional

from .exceptions import RevoteError, NotFullyImplementedWarning

from .api import HTTPClient
from .enums import Castle, CampLevel, World, Season, EventReservationOption, CrewRole, TShirtSize, SourcePoll
from .util import enum_converter, date_converter, character_converter, empty_string_converter


@attrs(repr=True, slots=True, frozen=True, hash=True)
class ParentSurveyResult:
    pass


@attrs(repr=True, slots=True, frozen=True, hash=True)
class ReservationManageDetails:
    pass


@attrs(repr=True, slots=True, eq=False)
class Resource:
    url = attrib(
        type=str,
        validator=instance_of(str)
    )
    _http = attrib(
        type=HTTPClient | None,
        default=None,
        validator=v_optional(
            instance_of(HTTPClient)
        ),
        repr=False
    )
    _cache_time = attrib(
        type=datetime | None,
        default=None,
        validator=v_optional(
            instance_of(datetime)
        ),
        kw_only=True,
        repr=False
    )
    _cache_content = attrib(
        type=bytes | None,
        default=None,
        kw_only=True,
        repr=False
    )

    async def get(self, use_cache: bool = True, update_cache: bool = True, http: HTTPClient | None = None) -> bytes:
        # Spoiler - Lock-i działają wolniej w tym przypadku...
        arguments = (self._cache_time, self._cache_content) if use_cache else ()
        client = http or self._http
        content = await client.get_resource(self.url, *arguments)
        if update_cache:
            self._cache_time = datetime.now()
            self._cache_content = content
        return content

    def __eq__(self, other: "Resource") -> bool:
        return self.url == other.url


@attrs(repr=True, slots=True, frozen=True, hash=True)
class Gallery:
    @attrs(repr=True, slots=True, frozen=True, hash=True)
    class Photo:
        normal = attrib(
            type=Resource,
            validator=instance_of(Resource)
        )
        large = attrib(
            type=Resource,
            validator=instance_of(Resource)
        )

        @classmethod
        def init_from_dict(cls, data: dict, **kwargs) -> "Photo":
            return cls(
                normal=Resource(data["AlbumUrl"], **kwargs),
                large=Resource(data["EnlargedUrl"], **kwargs)
            )

    gallery_id = attrib(
        type=int,
        validator=instance_of(int)
    )
    start = attrib(
        type=datetime | None,
        converter=c_optional(
            date_converter
        ),
        validator=v_optional(
            instance_of(datetime)
        ),
        default=None
    )
    end = attrib(
        type=datetime | None,
        converter=c_optional(
            date_converter
        ),
        validator=v_optional(
            instance_of(datetime)
        ),
        default=None
    )
    name = attrib(
        type=str | None,
        validator=v_optional(
            instance_of(str)
        ),
        default=None
    )
    empty = attrib(
        type=bool | None,
        validator=v_optional(
            instance_of(bool)
        ),
        default=None
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=v_optional(
            instance_of(HTTPClient),
        ),
        default=None,
        repr=False
    )

    async def get_photos(self, http: HTTPClient | None = None) -> List[Photo]:
        client = http or self._http
        photos = await client.get_images_galleries(self.gallery_id)
        return [
            self.Photo.init_from_dict(photo, http=client)
            for photo in photos
        ]

    @classmethod
    def init_from_dict(cls, data: dict, **kwargs) -> "Gallery":
        return cls(
            gallery_id=data["Id"],
            start=data["StartDate"],
            end=data["EndDate"],
            name=data["Name"],
            empty=not data["HasPhotos"],
            **kwargs
        )


@attrs(repr=True, slots=True, frozen=True, hash=True)
class Camp:
    @attrs(repr=True, slots=True, frozen=True, hash=True)
    class Transport:
        city = attrib(
            type=str,
            validator=instance_of(str)
        )
        one_way_price = attrib(
            type=int,
            validator=instance_of(int)
        )
        two_way_price = attrib(
            type=int,
            validator=instance_of(int)
        )

        @classmethod
        def init_from_dict(cls, data: dict) -> "Transport":
            return cls(
                city=data["City"],
                one_way_price=data["OneWayPrice"],
                two_way_price=data["TwoWayPrice"]
            )

    camp_id = attrib(
        type=int,
        validator=instance_of(int)
    )
    code = attrib(
        type=str,
        validator=instance_of(str)
    )
    place = attrib(
        type=Castle,
        converter=enum_converter(Castle),
        validator=instance_of(Castle)
    )
    price = attrib(
        type=int,
        validator=instance_of(int)
    )
    promo = attrib(
        type=int | None,
        validator=v_optional(
            instance_of(int)
        )
    )
    active = attrib(
        type=bool,
        validator=instance_of(bool)
    )
    places_left = attrib(
        type=int,
        validator=instance_of(int)
    )
    program = attrib(
        type=str,
        validator=instance_of(str)
    )
    level = attrib(
        type=CampLevel,
        converter=enum_converter(CampLevel),
        validator=instance_of(CampLevel)
    )
    world = attrib(
        type=World,
        converter=enum_converter(World),
        validator=instance_of(World)
    )
    season = attrib(
        type=Season,
        converter=enum_converter(Season),
        validator=instance_of(Season)
    )
    trip = attrib(
        type=str | None,
        converter=empty_string_converter,
        validator=v_optional(
            instance_of(str)
        )
    )
    start = attrib(
        type=datetime,
        converter=lambda value: value if isinstance(value, datetime) else date_converter(value),
        validator=instance_of(datetime)
    )
    end = attrib(
        type=datetime,
        converter=lambda value: value if isinstance(value, datetime) else date_converter(value),
        validator=instance_of(datetime)
    )
    ages = attrib(
        type=List[str],
        validator=deep_iterable(
            instance_of(str)
        )
    )
    transports = attrib(
        type=List[Transport],
        validator=deep_iterable(
            instance_of(Transport)
        )
    )

    @classmethod
    def init_from_dict(cls, data: dict) -> "Camp":
        return cls(
            data["Id"],
            data["Code"],
            data["Place"],
            data["Price"],
            data["Promo"],
            data["IsActive"],
            data["PlacesLeft"],
            data["Program"],
            data["Level"],
            data["World"],
            data["Season"],
            data["Trip"],
            data["StartDate"],
            data["EndDate"],
            data["Ages"],
            [
                cls.Transport.init_from_dict(transport)
                for transport in data["Transports"]
            ]
        )


@attrs(repr=True, slots=True, frozen=True, hash=True)
class Purchaser:
    name = attrib(
        type=str,
        validator=instance_of(str)
    )
    surname = attrib(
        type=str,
        validator=instance_of(str)
    )
    email = attrib(
        type=str,
        validator=instance_of(str)
    )
    phone = attrib(
        type=str,
        validator=instance_of(str)
    )
    parcel_locker = attrib(
        type=str,
        validator=instance_of(str)
    )

    def to_dict(self) -> dict:
        return {
            "Name": self.name,
            "Surname": self.surname,
            "Email": self.email,
            "Phone": self.phone,
            "ParcelLocker": self.parcel_locker
        }


@attrs(repr=True, slots=True, frozen=True, hash=True)
class PersonalReservationInfo:
    reservation_id = attrib(
        type=str,
        validator=instance_of(str)
    )
    surname = attrib(
        type=str,
        validator=instance_of(str)
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=instance_of(HTTPClient),
        default=None,
        repr=False
    )

    def to_dict(self) -> dict:
        return {
            "ReservationId": self.reservation_id,
            "Surname": self.surname
        }

    async def get_details(self, http: HTTPClient | None) -> ReservationManageDetails | dict:
        warn(
            "Ta metoda będzie w przyszłości zwracała ReservationMangeDetails."
            "Jeśli chcesz pomóc w jej implementacji otwórz nowy issue: https://github.com/Asapros/pymondis/issues",
            NotFullyImplementedWarning
        )
        client = http or self._http
        details = await client.post_reservations_manage(self.to_dict())
        return details


@attrs(repr=True, slots=True, frozen=True, hash=True)
class WebReservationModel:
    class Child:
        name = attrib(
            type=str,
            validator=instance_of(str)
        )
        surname = attrib(
            type=str,
            validator=instance_of(str)
        )
        t_shirt_size = attrib(
            type=TShirtSize,
            converter=enum_converter(TShirtSize),
            validator=instance_of(TShirtSize)
        )
        birthdate = attrib(
            type=datetime,
            converter=date_converter,
            validator=instance_of(datetime)
        )

        def to_dict(self) -> dict:
            return {
                "Name": self.name,
                "Surname": self.surname,
                "Tshirt": self.t_shirt_size.value,
                "Dob": self.birthdate
            }

    camp_id = attrib(
        type=int,
        validator=instance_of(int)
    )
    child = attrib(
        type=Child,
        validator=instance_of(Child)
    )
    parent_name = attrib(
        type=str,
        validator=instance_of(str)
    )
    parent_surname = attrib(
        type=str,
        validator=instance_of(str)
    )
    nip = attrib(
        type=str,
        validator=instance_of(str)
    )
    email = attrib(
        type=str,
        validator=instance_of(str)
    )
    phone = attrib(
        type=str,
        validator=instance_of(str)
    )
    poll = attrib(
        type=SourcePoll,
        converter=enum_converter(SourcePoll),
        validator=instance_of(SourcePoll)
    )
    siblings = attrib(
        type=List[Child],
        validator=deep_iterable(
            instance_of(Child)
        ),
        factory=list
    )
    promo_code = attrib(
        type=str | None,
        validator=v_optional(
            instance_of(str)
        ),
        default=None
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=v_optional(
            instance_of(HTTPClient)
        ),
        default=None,
        repr=False
    )

    def to_dict(self) -> dict:
        return {
            "SubcampId": self.camp_id,
            "Childs": {  # English 100
                "Main": self.child.to_dict(),
                "Siblings": [sibling.to_dict() for sibling in self.siblings]
            },
            "Parent": {
                "Name": self.parent_name,
                "Surname": self.parent_surname,
                "Nip": self.nip
            },
            "Details": {
                "Email": self.email,
                "Phone": self.phone,
                "Promo": self.promo_code,
                "Poll": self.poll.value
            }
        }

    @property
    def pri(self, **kwargs) -> PersonalReservationInfo:
        kwargs = {"http": self._http} | kwargs
        return PersonalReservationInfo(self.camp_id, self.parent_surname, **kwargs)


@attrs(repr=True, slots=True, frozen=True, hash=True)
class EventReservationSummary:
    option = attrib(
        type=EventReservationOption,
        converter=enum_converter(EventReservationOption),
        validator=instance_of(EventReservationOption)
    )
    name = attrib(
        type=str,
        validator=instance_of(str)
    )
    surname = attrib(
        type=str,
        validator=instance_of(str)
    )
    parent_name = attrib(
        type=str,
        validator=instance_of(str)
    )
    parent_surname = attrib(
        type=str,
        validator=instance_of(str)
    )
    parent_reused = attrib(
        type=bool,
        validator=instance_of(bool)
    )
    phone = attrib(
        type=str,
        validator=instance_of(str)
    )
    email = attrib(
        type=str,
        validator=instance_of(str)
    )
    first_parent_name = attrib(
        type=str | None,
        validator=v_optional(
            instance_of(str)
        )
    )
    first_parent_surname = attrib(
        type=str | None,
        validator=v_optional(
            instance_of(str)
        )
    )
    second_parent_name = attrib(
        type=str | None,
        validator=v_optional(
            instance_of(str)
        )
    )
    second_parent_surname = attrib(
        type=str | None,
        validator=v_optional(
            instance_of(str)
        )
    )
    _price = attrib(
        type=int,
        validator=v_optional(
            instance_of(int)
        ),
        kw_only=True
    )

    @property
    def price(self) -> int:
        if self._price is None:
            match self.option:
                case EventReservationOption.CHILD:
                    return 450
                case EventReservationOption.CHILD_AND_ONE_PARENT:
                    return 900
                case EventReservationOption.CHILD_AND_TWO_PARENTS:
                    return 1300
            raise ValueError("Option is not one of the enum elements")
        return self._price

    def to_dict(self) -> dict:
        data = {
            "Price": self.price,
            "Name": self.name,
            "Surname": self.surname,
            "ParentName": self.parent_name,
            "ParentSurname": self.parent_surname,
            "IsParentReused": self.parent_reused,
            "Phone": self.phone,
            "Email": self.email
        }
        if self.option in (EventReservationOption.CHILD, EventReservationOption.CHILD_AND_PARENT):
            data.update({"FirstParentName": self.first_parent_name, "FirstParentSurname": self.first_parent_surname})
        if self.option is EventReservationOption.CHILD_AND_TWO_PARENTS:
            data.update(
                {"SecondParentName": self.second_parent_name, "SecondParentSurname": self.second_parent_surname})
        return data


@attrs(repr=True, slots=True, frozen=True, hash=True)
class CrewMember:
    name = attrib(
        type=str,
        validator=instance_of(str)
    )
    surname = attrib(
        type=str,
        validator=instance_of(str)
    )
    character = attrib(
        type=str | None,
        converter=character_converter,
        validator=v_optional(
            instance_of(str)
        )
    )
    position = attrib(
        type=CrewRole,
        converter=enum_converter(CrewRole),
        validator=instance_of(CrewRole)
    )
    description = attrib(
        type=str,
        validator=instance_of(str)
    )
    photo = attrib(
        type=Resource,
        validator=instance_of(Resource)
    )

    @classmethod
    def init_from_dict(cls, data: dict, **kwargs) -> "CrewMember":
        return cls(
            name=data["Name"],
            surname=data["Surname"],
            character=data["Character"].strip(),
            position=data["Position"],
            description=data["Description"],
            photo=Resource(data["PhotoUrl"], **kwargs)
        )


@attrs(repr=True, slots=True, frozen=True, hash=True)
class PlebisciteCandidate:
    name = attrib(
        type=str,
        validator=instance_of(str)
    )
    category = attrib(
        type=str,
        validator=instance_of(str)
    )
    votes = attrib(
        type=int | None,
        validator=v_optional(
            instance_of(int)
        ),
        default=None
    )
    plebiscite = attrib(
        type=str | None,
        validator=v_optional(
            instance_of(str)
        ),
        default=None
    )
    voted = attrib(
        type=bool | None,
        validator=v_optional(
            instance_of(bool)
        ),
        default=None
    )
    _http = attrib(
        type=HTTPClient | None,
        validator=v_optional(
            instance_of(HTTPClient)
        ),
        default=None,
        repr=False
    )

    @classmethod
    def init_from_dict(cls, data: dict, **kwargs) -> "PlebisciteCandidate":
        return cls(
            name=data["Name"],
            votes=data["Result"],
            category=data["Category"],
            plebiscite=data["Plebiscite"],
            voted=data["WasVoted"],
            **kwargs
        )

    async def vote(self, http: HTTPClient | None = None):
        if self.voted:
            raise RevoteError(self.category)
        client = http or self._http
        await client.patch_vote(self.category, self.name)


Photo = Gallery.Photo
Transport = Camp.Transport
Child = WebReservationModel.Child
